@@ -41,10 +41,12 @@
 #include "input.h"
 #include "parser.h"	/* for the struct dstack stuff. */
 #include "pathexp.h"	/* for the struct ignorevar stuff */
+#include "bashhist.h"	/* matching prototypes and declarations */
 #include "builtins/common.h"
 
 #include <readline/history.h>
-#include <glob/fnmatch.h>
+#include <glob/glob.h>
+#include <glob/strmatch.h>
 
 #if defined (READLINE)
 #  include "bashline.h"
@@ -54,9 +56,9 @@
 extern int errno;
 #endif
 
-extern int glob_pattern_p ();
-
-static int histignore_item_func ();
+static int histignore_item_func __P((struct ign *));
+static int check_history_control __P((char *));
+static void really_add_history __P((char *));
 
 static struct ignorevar histignore =
 {
@@ -64,7 +66,7 @@ static struct ignorevar histignore =
   (struct ign *)0,
   0,
   (char *)0,
-  (Function *)histignore_item_func,
+  (sh_iv_item_func_t *)histignore_item_func,
 };
 
 #define HIGN_EXPAND 0x01
@@ -87,8 +89,8 @@ int history_lines_in_file;
 int history_expansion_inhibited;
 #endif
 
-/* By default, every line is saved in the history individually.  I.e.,
-   if the user enters:
+/* With the old default, every line was saved in the history individually.
+   I.e., if the user enters:
 	bash$ for i in a b c
 	> do
 	> echo $i
@@ -114,9 +116,16 @@ int history_expansion_inhibited;
 	11  history
    The user can then recall the whole command all at once instead
    of just being able to recall one line at a time.
+
+   This is now enabled by default.
    */
 int command_oriented_history = 1;
 
+/* Set to 1 if the first line of a possibly-multi-line command was saved
+   in the history list.  Managed by maybe_add_history(), but global so
+   the history-manipluating builtins can see it. */
+int current_command_first_line_saved = 0;
+
 /* Non-zero means to store newlines in the history list when using
    command_oriented_history rather than trying to use semicolons. */
 int literal_history;
@@ -148,18 +157,24 @@ int hist_verify;
 
 #endif /* READLINE */
 
+/* Non-zero means to not save function definitions in the history list. */
+int dont_save_function_defs;
+
 /* Variables declared in other files used here. */
-extern int interactive;
 extern int current_command_line_count;
 
 extern struct dstack dstack;
 
-extern char *extract_colon_unit ();
-extern char *history_delimiting_chars ();
-extern void maybe_add_history ();	/* forward declaration */
-extern void bash_add_history ();	/* forward declaration */
-
-static int history_should_ignore ();
+static int bash_history_inhibit_expansion __P((char *, int));
+#if defined (READLINE)
+static void re_edit __P((char *));
+#endif
+static int history_expansion_p __P((char *));
+static int shell_comment __P((char *));
+static int should_expand __P((char *));
+static HIST_ENTRY *last_history_entry __P((void));
+static char *expand_histignore_pattern __P((char *));
+static int history_should_ignore __P((char *));
 
 /* Is the history expansion starting at string[i] one that should not
    be expanded? */
@@ -470,6 +485,7 @@ shell_comment (line)
   return (p && *p == '#');
 }
 
+#ifdef INCLUDE_UNUSED
 /* Remove shell comments from LINE.  A `#' and anything after it is a comment.
    This isn't really useful yet, since it doesn't handle quoting. */
 static char *
@@ -484,57 +500,93 @@ filter_comments (line)
     *p = '\0';
   return (line);
 }
+#endif
 
-/* Add LINE to the history list depending on the value of HISTORY_CONTROL. */
+/* Check LINE against what HISTCONTROL says to do.  Returns 1 if the line
+   should be saved; 0 if it should be discarded. */
+static int
+check_history_control (line)
+     char *line;
+{
+  HIST_ENTRY *temp;
+  int r;
+
+  switch (history_control)
+    {
+    case 0:			/* nothing */
+      return 1;
+    case 1:			/* ignorespace */
+      return (*line != ' ');
+    case 3:			/* ignoreboth */
+      if (*line == ' ')
+	return 0;
+      /* FALLTHROUGH if case == 3 (`ignoreboth') */
+    case 2:			/* ignoredups */
+      using_history ();
+      temp = previous_history ();
+
+      r = (temp == 0 || STREQ (temp->line, line) == 0);
+
+      using_history ();
+      return r;
+    }
+
+  return 0;
+}
+
+/* Add LINE to the history list, handling possibly multi-line compound
+   commands.  We note whether or not we save the first line of each command
+   (which is usually the entire command and history entry), and don't add
+   the second and subsequent lines of a multi-line compound command if we
+   didn't save the first line.  We don't usually save shell comment lines in
+   compound commands in the history, because they could have the effect of
+   commenting out the rest of the command when the entire command is saved as
+   a single history entry (when COMMAND_ORIENTED_HISTORY is enabled).  If
+   LITERAL_HISTORY is set, we're saving lines in the history with embedded
+   newlines, so it's OK to save comment lines.  We also make sure to save
+   multiple-line quoted strings or other constructs. */
 void
 maybe_add_history (line)
      char *line;
 {
-  int should_add;
-  HIST_ENTRY *temp;
-
-  should_add = hist_last_line_added = 0;
+  hist_last_line_added = 0;
 
   /* Don't use the value of history_control to affect the second
      and subsequent lines of a multi-line command (old code did
      this only when command_oriented_history is enabled). */
-#if 0
-  if (command_oriented_history && current_command_line_count > 1)
-#else
   if (current_command_line_count > 1)
-#endif
     {
-      if (literal_history || dstack.delimiter_depth != 0 || shell_comment (line) == 0)
+      if (current_command_first_line_saved &&
+	  (literal_history || dstack.delimiter_depth != 0 || shell_comment (line) == 0))
 	bash_add_history (line);
       return;
     }
 
-  switch (history_control)
+  /* This is the first line of a (possible multi-line) command.  Note whether
+     or not we should save the first line and remember it. */
+  current_command_first_line_saved = check_add_history (line, 0);
+}
+
+/* Just check LINE against HISTCONTROL and HISTIGNORE and add it to the
+   history if it's OK.  Used by `history -s' as well as maybe_add_history().
+   Returns 1 if the line was saved in the history, 0 otherwise. */
+int
+check_add_history (line, force)
+     char *line;
+     int force;
+{
+  if (check_history_control (line) && history_should_ignore (line) == 0)
     {
-    case 0:
-      should_add = 1;
-      break;
-    case 1:
-      if (*line != ' ')
-	should_add = 1;
-      break;
-    case 3:
-      if (*line == ' ')
-	break;
-      /* FALLTHROUGH if case == 3 (`ignoreboth') */
-    case 2:
-      using_history ();
-      temp = previous_history ();
-
-      if (temp == 0 || STREQ (temp->line, line) == 0)
-	should_add = 1;
-
-      using_history ();
-      break;
+      if (force)
+	{
+	  really_add_history (line);
+	  using_history ();
+	}
+      else
+	bash_add_history (line);
+      return 1;
     }
-
-  if (should_add && history_should_ignore (line) == 0)
-    bash_add_history (line);
+  return 0;
 }
 
 /* Add a line to the history list.
@@ -573,10 +625,10 @@ bash_add_history (line)
 	      chars_to_add = "";
 	    }
 
-	  new_line = (char *) xmalloc (1
-				       + curlen
-				       + strlen (line)
-				       + strlen (chars_to_add));
+	  new_line = (char *)xmalloc (1
+				      + curlen
+				      + strlen (line)
+				      + strlen (chars_to_add));
 	  sprintf (new_line, "%s%s%s", current->line, chars_to_add, line);
 	  offset = where_history ();
 	  old = replace_history_entry (offset, new_line, current->data);
@@ -592,14 +644,20 @@ bash_add_history (line)
     }
 
   if (add_it)
-    {
-      hist_last_line_added = 1;
-      add_history (line);
-      history_lines_this_session++;
-    }
+    really_add_history (line);
+
   using_history ();
 }
 
+static void
+really_add_history (line)
+     char *line;
+{
+  hist_last_line_added = 1;
+  add_history (line);
+  history_lines_this_session++;
+}
+
 int
 history_number ()
 {
@@ -697,7 +755,7 @@ history_should_ignore (line)
       else
 	npat = histignore.ignores[i].val;
 
-      match = fnmatch (npat, line, FNMATCH_EXTFLAG) != FNM_NOMATCH;
+      match = strmatch (npat, line, FNMATCH_EXTFLAG) != FNM_NOMATCH;
 
       if (histignore.ignores[i].flags & HIGN_EXPAND)
 	free (npat);
