@@ -6,7 +6,7 @@ This file is part of GNU Bash, the Bourne Again SHell.
 
 Bash is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free
-Software Foundation; either version 1, or (at your option) any later
+Software Foundation; either version 2, or (at your option) any later
 version.
 
 Bash is distributed in the hope that it will be useful, but WITHOUT ANY
@@ -16,7 +16,7 @@ for more details.
 
 You should have received a copy of the GNU General Public License along
 with Bash; see the file COPYING.  If not, write to the Free Software
-Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */
+Foundation, 59 Temple Place, Suite 330, Boston, MA 02111 USA. */
 
 #include <config.h>
 
@@ -25,6 +25,7 @@ Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */
 #endif
 
 #include "../bashansi.h"
+#include <chartypes.h>
 #include <errno.h>
 
 #include "../shell.h"
@@ -124,7 +125,7 @@ char		*opts;
 	} else if (*cp == '#') {
 		/* optional numeric argument */
 		if (lcurrent->word->word[sp+1]) {
-			if (digit(lcurrent->word->word[sp+1])) {
+			if (DIGIT(lcurrent->word->word[sp+1])) {
 				list_optarg = lcurrent->word->word + sp + 1;
 				lcurrent = lcurrent->next;
 			} else
