@@ -8,7 +8,7 @@
 
    Bash is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 1, or (at your option)
+   the Free Software Foundation; either version 2, or (at your option)
    any later version.
 
    Bash is distributed in the hope that it will be useful, but WITHOUT
@@ -18,7 +18,7 @@
 
    You should have received a copy of the GNU General Public License
    along with Bash; see the file COPYING.  If not, write to the Free
-   Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */
+   Software Foundation, 59 Temple Place, Suite 330, Boston, MA 02111 USA. */
 
 #include "config.h"
 
@@ -32,6 +32,26 @@
 
 #include "shell.h"
 
+static PATTERN_LIST *copy_case_clause __P((PATTERN_LIST *));
+static PATTERN_LIST *copy_case_clauses __P((PATTERN_LIST *));
+static FOR_COM *copy_for_command __P((FOR_COM *));
+#if defined (ARITH_FOR_COMMAND)
+static ARITH_FOR_COM *copy_arith_for_command __P((ARITH_FOR_COM *));
+#endif
+static GROUP_COM *copy_group_command __P((GROUP_COM *));
+static SUBSHELL_COM *copy_subshell_command __P((SUBSHELL_COM *));
+static CASE_COM *copy_case_command __P((CASE_COM *));
+static WHILE_COM *copy_while_command __P((WHILE_COM *));
+static IF_COM *copy_if_command __P((IF_COM *));
+#if defined (DPAREN_ARITHMETIC)
+static ARITH_COM *copy_arith_command __P((ARITH_COM *));
+#endif
+#if defined (COND_COMMAND)
+static COND_COM *copy_cond_command __P((COND_COM *));
+#endif
+static SIMPLE_COM *copy_simple_command __P((SIMPLE_COM *));
+static FUNCTION_DEF *copy_function_def __P((FUNCTION_DEF *));
+
 WORD_DESC *
 copy_word (w)
      WORD_DESC *w;
@@ -39,7 +59,11 @@ copy_word (w)
   WORD_DESC *new_word;
 
   new_word = (WORD_DESC *)xmalloc (sizeof (WORD_DESC));
+#if 1
+  new_word->flags = w->flags;
+#else
   FASTCOPY ((char *)w, (char *)new_word, sizeof (WORD_DESC));
+#endif
   new_word->word = savestring (w->word);
   return (new_word);
 }
@@ -152,6 +176,24 @@ copy_for_command (com)
   return (new_for);
 }
 
+#if defined (ARITH_FOR_COMMAND)
+static ARITH_FOR_COM *
+copy_arith_for_command (com)
+     ARITH_FOR_COM *com;
+{
+  ARITH_FOR_COM *new_arith_for;
+
+  new_arith_for = (ARITH_FOR_COM *)xmalloc (sizeof (ARITH_FOR_COM));
+  new_arith_for->flags = com->flags;
+  new_arith_for->line = com->line;
+  new_arith_for->init = copy_word_list (com->init);
+  new_arith_for->test = copy_word_list (com->test);
+  new_arith_for->step = copy_word_list (com->step);
+  new_arith_for->action = copy_command (com->action);
+  return (new_arith_for);
+}
+#endif /* ARITH_FOR_COMMAND */
+
 static GROUP_COM *
 copy_group_command (com)
      GROUP_COM *com;
@@ -163,6 +205,18 @@ copy_group_command (com)
   return (new_group);
 }
 
+static SUBSHELL_COM *
+copy_subshell_command (com)
+     SUBSHELL_COM *com;
+{
+  SUBSHELL_COM *new_subshell;
+
+  new_subshell = (SUBSHELL_COM *)xmalloc (sizeof (SUBSHELL_COM));
+  new_subshell->command = copy_command (com->command);
+  new_subshell->flags = com->flags;
+  return (new_subshell);
+}
+
 static CASE_COM *
 copy_case_command (com)
      CASE_COM *com;
@@ -261,6 +315,8 @@ copy_function_def (com)
   new_def = (FUNCTION_DEF *)xmalloc (sizeof (FUNCTION_DEF));
   new_def->name = copy_word (com->name);
   new_def->command = copy_command (com->command);
+  new_def->flags = com->flags;
+  new_def->line = com->line;
   return (new_def);
 }
 
@@ -290,6 +346,12 @@ copy_command (command)
 	new_command->value.For = copy_for_command (command->value.For);
 	break;
 
+#if defined (ARITH_FOR_COMMAND)
+      case cm_arith_for:
+	new_command->value.ArithFor = copy_arith_for_command (command->value.ArithFor);
+	break;
+#endif
+
 #if defined (SELECT_COMMAND)
       case cm_select:
 	new_command->value.Select =
@@ -301,6 +363,10 @@ copy_command (command)
 	new_command->value.Group = copy_group_command (command->value.Group);
 	break;
 
+      case cm_subshell:
+	new_command->value.Subshell = copy_subshell_command (command->value.Subshell);
+	break;
+
       case cm_case:
 	new_command->value.Case = copy_case_command (command->value.Case);
 	break;
@@ -316,8 +382,8 @@ copy_command (command)
 
 #if defined (DPAREN_ARITHMETIC)
       case cm_arith:
-        new_command->value.Arith = copy_arith_command (command->value.Arith);
-        break;
+	new_command->value.Arith = copy_arith_command (command->value.Arith);
+	break;
 #endif
 
 #if defined (COND_COMMAND)
