@@ -6,7 +6,7 @@
 
    Bash is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 1, or (at your option)
+   the Free Software Foundation; either version 2, or (at your option)
    any later version.
 
    Bash is distributed in the hope that it will be useful, but WITHOUT
@@ -16,7 +16,7 @@
 
    You should have received a copy of the GNU General Public License
    along with Bash; see the file COPYING.  If not, write to the Free
-   Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */
+   Software Foundation, 59 Temple Place, Suite 330, Boston, MA 02111 USA. */
 
 /*
  All arithmetic is done as long integers with no checking for overflow
@@ -25,6 +25,8 @@
  The following operators are handled, grouped into a set of levels in
  order of decreasing precedence.
 
+	"id++", "id--"		[post-increment and post-decrement]
+	"++id", "--id"		[pre-increment and pre-decrement]
 	"-", "+"		[(unary operators)]
 	"!", "~"
 	"**"			[(exponentiation)]
@@ -75,6 +77,8 @@
 #  include <unistd.h>
 #endif
 
+#include "chartypes.h"
+
 #include "shell.h"
 
 /* Because of the $((...)) construct, expressions may include newlines.
@@ -104,6 +108,10 @@
 #define OP_ASSIGN 11	/* op= expassign as in Posix.2 */
 #define COND	12	/* exp1 ? exp2 : exp3 */
 #define POWER	13	/* exp1**exp2 */
+#define PREINC	14	/* ++var */
+#define PREDEC	15	/* --var */
+#define POSTINC	16	/* var++ */
+#define POSTDEC	17	/* var-- */
 #define EQ	'='
 #define GT	'>'
 #define LT	'<'
@@ -121,6 +129,11 @@
 #define BNOT	'~'	/* Bitwise NOT; Two's complement. */
 #define QUES	'?'
 #define COL	':'
+#define COMMA	','
+
+/* This should be the function corresponding to the operator with the
+   highest precedence. */
+#define EXP_HIGHEST	expcomma
 
 static char	*expression;	/* The current expression */
 static char	*tp;		/* token lexical position */
@@ -129,25 +142,53 @@ static int	curtok;		/* the current token */
 static int	lasttok;	/* the previous token */
 static int	assigntok;	/* the OP in OP= */
 static char	*tokstr;	/* current token string */
-static int	tokval;		/* current token value */
+static long	tokval;		/* current token value */
 static int	noeval;		/* set to 1 if no assignment to be done */
 static procenv_t evalbuf;
 
-static void	readtok ();	/* lexical analyzer */
-static long	subexpr (), expassign (), exp0 (), exp1 (), exp2 (), exp3 (),
-		exp4 (), exp5 (), expshift (), expland (), explor (),
-		expband (), expbor (), expbxor (), expcond (), exppower ();
-static long	strlong ();
-static void	evalerror ();
+static void	readtok __P((void));	/* lexical analyzer */
+static long	strlong __P((char *));
+static void	evalerror __P((char *));
+
+static void	pushexp __P((void));
+static void	popexp __P((void));
+
+static long 	subexpr __P((char *));
+
+static long	expcomma __P((void));
+static long 	expassign __P((void));
+static long	expcond __P((void));
+static long 	explor __P((void));
+static long 	expland __P((void));
+static long	expbor __P((void));
+static long	expbxor __P((void));
+static long	expband __P((void));
+static long 	exp5 __P((void));
+static long 	exp4 __P((void));
+static long 	expshift __P((void));
+static long 	exp3 __P((void));
+static long 	exp2 __P((void));
+static long	exppower __P((void));
+static long 	exp1 __P((void));
+static long 	exp0 __P((void));
 
 /* A structure defining a single expression context. */
 typedef struct {
   int curtok, lasttok;
-  char *expression, *tp;
-  int tokval;
+  char *expression, *tp, *lasttp;
+  long tokval;
   char *tokstr;
+  int noeval;
 } EXPR_CONTEXT;
 
+#ifdef INCLUDE_UNUSED
+/* Not used yet. */
+typedef struct {
+  char *tokstr;
+  long tokval;
+} LVALUE;
+#endif
+
 /* Global var which contains the stack of expression contexts. */
 static EXPR_CONTEXT **expr_stack;
 static int expr_depth;		   /* Location in the stack. */
@@ -155,6 +196,28 @@ static int expr_stack_size;	   /* Number of slots already allocated. */
 
 extern char *this_command_name;
 
+#define SAVETOK(X) \
+  do { \
+    (X)->curtok = curtok; \
+    (X)->lasttok = lasttok; \
+    (X)->tp = tp; \
+    (X)->lasttp = lasttp; \
+    (X)->tokval = tokval; \
+    (X)->tokstr = tokstr; \
+    (X)->noeval = noeval; \
+  } while (0)
+
+#define RESTORETOK(X) \
+  do { \
+    curtok = (X)->curtok; \
+    lasttok = (X)->lasttok; \
+    tp = (X)->tp; \
+    lasttp = (X)->lasttp; \
+    tokval = (X)->tokval; \
+    tokstr = (X)->tokstr; \
+    noeval = (X)->noeval; \
+  } while (0)
+
 /* Push and save away the contents of the globals describing the
    current expression context. */
 static void
@@ -167,19 +230,15 @@ pushexp ()
 
   if (expr_depth >= expr_stack_size)
     {
-      expr_stack = (EXPR_CONTEXT **)
-	xrealloc (expr_stack, (expr_stack_size += EXPR_STACK_GROW_SIZE)
-		  * sizeof (EXPR_CONTEXT *));
+      expr_stack_size += EXPR_STACK_GROW_SIZE;
+      expr_stack = (EXPR_CONTEXT **)xrealloc (expr_stack, expr_stack_size * sizeof (EXPR_CONTEXT *));
     }
 
   context = (EXPR_CONTEXT *)xmalloc (sizeof (EXPR_CONTEXT));
 
-  context->curtok = curtok;
-  context->lasttok = lasttok;
   context->expression = expression;
-  context->tp = tp;
-  context->tokval = tokval;
-  context->tokstr = tokstr;
+  SAVETOK(context);
+
   expr_stack[expr_depth++] = context;
 }
 
@@ -194,12 +253,10 @@ popexp ()
     evalerror ("recursion stack underflow");
 
   context = expr_stack[--expr_depth];
-  curtok = context->curtok;
-  lasttok = context->lasttok;
+
   expression = context->expression;
-  tp = context->tp;
-  tokval = context->tokval;
-  tokstr = context->tokstr;
+  RESTORETOK (context);
+
   free (context);
 }
 
@@ -226,7 +283,7 @@ evalexp (expr, validp)
   procenv_t old_evalbuf;
 #endif
 
-  val = 0L;
+  val = 0;
 
 #if 0
   /* Save the value of evalbuf to protect it around possible recursive
@@ -254,7 +311,7 @@ evalexp (expr, validp)
 
       if (validp)
 	*validp = 0;
-      return (0L);
+      return (0);
     }
 
   val = subexpr (expr);
@@ -282,7 +339,7 @@ subexpr (expr)
     ;
 
   if (p == NULL || *p == '\0')
-    return (0L);
+    return (0);
 
   pushexp ();
   curtok = lasttok = 0;
@@ -290,11 +347,11 @@ subexpr (expr)
   tp = expression;
 
   tokstr = (char *)NULL;
-  tokval = 0L;
+  tokval = 0;
 
   readtok ();
 
-  val = expassign ();
+  val = EXP_HIGHEST ();
 
   if (curtok != 0)
     evalerror ("syntax error in expression");
@@ -307,35 +364,21 @@ subexpr (expr)
   return val;
 }
 
-/* Bind/create a shell variable with the name LHS to the RHS.
-   This creates or modifies a variable such that it is an integer.
-
-   This should really be in variables.c, but it is here so that all of the
-   expression evaluation stuff is localized.  Since we don't want any
-   recursive evaluation from bind_variable() (possible without this code,
-   since bind_variable() calls the evaluator for variables with the integer
-   attribute set), we temporarily turn off the integer attribute for each
-   variable we set here, then turn it back on after binding as necessary. */
-
-void
-bind_int_variable (lhs, rhs)
-     char *lhs, *rhs;
+static long
+expcomma ()
 {
-  register SHELL_VAR *v;
-  int isint = 0;
+  register long value;
 
-  v = find_variable (lhs);
-  if (v)
+  value = expassign ();
+  while (curtok == COMMA)
     {
-      isint = integer_p (v);
-      v->attributes &= ~att_integer;
+      readtok ();
+      value = expassign ();
     }
 
-  v = bind_variable (lhs, rhs);
-  if (isint)
-    v->attributes |= att_integer;
+  return value;
 }
-
+  
 static long
 expassign ()
 {
@@ -394,6 +437,9 @@ expassign ()
 	    case BOR:
 	      lvalue |= value;
 	      break;
+	    case BXOR:
+	      lvalue ^= value;
+	      break;
 	    default:
 	      free (lhs);
 	      evalerror ("bug: bad expassign token");
@@ -404,7 +450,7 @@ expassign ()
 
       rhs = itos (value);
       if (noeval == 0)
-	bind_int_variable (lhs, rhs);
+	(void)bind_int_variable (lhs, rhs);
       free (rhs);
       free (lhs);
       FREE (tokstr);
@@ -432,15 +478,13 @@ expcond ()
 	  set_noeval = 1;
 	  noeval++;
 	}
-#if 0
-      val1 = explor ();
-#else
-      val1 = expassign ();
-#endif
+
+      val1 = EXP_HIGHEST ();
+
       if (set_noeval)
-        noeval--;
+	noeval--;
       if (curtok != COL)
-        evalerror ("`:' expected for conditional expression");
+	evalerror ("`:' expected for conditional expression");
       readtok ();
       if (curtok == 0)
 	evalerror ("expression expected");
@@ -452,7 +496,7 @@ expcond ()
  	}
       val2 = explor ();
       if (set_noeval)
-        noeval--;
+	noeval--;
       rval = cval ? val1 : val2;
       lasttok = COND;
     }
@@ -672,8 +716,8 @@ exp2 ()
   val1 = exppower ();
 
   while ((curtok == MUL) ||
-         (curtok == DIV) ||
-         (curtok == MOD))
+	 (curtok == DIV) ||
+	 (curtok == MOD))
     {
       int op = curtok;
 
@@ -685,11 +729,11 @@ exp2 ()
 	evalerror ("division by 0");
 
       if (op == MUL)
-        val1 *= val2;
+	val1 *= val2;
       else if (op == DIV)
-        val1 /= val2;
+	val1 /= val2;
       else if (op == MOD)
-        val1 %= val2;
+	val1 %= val2;
     }
   return (val1);
 }
@@ -705,7 +749,9 @@ exppower ()
       readtok ();
       val2 = exp1 ();
       if (val2 == 0)
-	return (1L);
+	return (1);
+      if (val2 < 0)
+	evalerror ("exponent less than 0");
       for (c = 1; val2--; c *= val1)
 	;
       val1 = c;
@@ -737,9 +783,31 @@ exp1 ()
 static long
 exp0 ()
 {
-  register long val = 0L;
+  register long val = 0, v2;
+  char *vincdec;
+  int stok;
 
-  if (curtok == MINUS)
+  /* XXX - might need additional logic here to decide whether or not
+	   pre-increment or pre-decrement is legal at this point. */
+  if (curtok == PREINC || curtok == PREDEC)
+    {
+      stok = lasttok = curtok;
+      readtok ();
+      if (curtok != STR)
+	/* readtok() catches this */
+	evalerror ("identifier expected after pre-increment or pre-decrement");
+
+      v2 = tokval + ((stok == PREINC) ? 1 : -1);
+      vincdec = itos (v2);
+      if (noeval == 0)
+	(void)bind_int_variable (tokstr, vincdec);
+      free (vincdec);
+      val = v2;
+
+      curtok = NUM;	/* make sure --x=7 is flagged as an error */
+      readtok ();
+    }
+  else if (curtok == MINUS)
     {
       readtok ();
       val = - exp0 ();
@@ -752,7 +820,7 @@ exp0 ()
   else if (curtok == LPAR)
     {
       readtok ();
-      val = expassign ();
+      val = EXP_HIGHEST ();
 
       if (curtok != RPAR)
 	evalerror ("missing `)'");
@@ -763,6 +831,19 @@ exp0 ()
   else if ((curtok == NUM) || (curtok == STR))
     {
       val = tokval;
+      if (curtok == STR && (*tp == '+' || *tp == '-') && tp[1] == *tp &&
+		(tp[2] == '\0' || (ISALNUM ((unsigned char)tp[2]) == 0)))
+	{
+	  /* post-increment or post-decrement */
+	  v2 = val + ((*tp == '+') ? 1 : -1);
+	  vincdec = itos (v2);
+	  if (noeval == 0)
+	    (void)bind_int_variable (tokstr, vincdec);
+	  free (vincdec);
+	  tp += 2;
+	  curtok = NUM;	/* make sure x++=7 is flagged as an error */
+	}
+	  
       readtok ();
     }
   else
@@ -779,7 +860,8 @@ static void
 readtok ()
 {
   register char *cp;
-  register int c, c1, e;
+  register unsigned char c, c1;
+  register int e;
 
   /* Skip leading whitespace. */
   cp = tp;
@@ -802,9 +884,10 @@ readtok ()
 
   if (legal_variable_starter (c))
     {
-      /* Semi-bogus ksh compatibility feature -- variable names
-	 not preceded with a dollar sign are shell variables. */
-      char *value;
+      /* variable names not preceded with a dollar sign are shell variables. */
+      char *value, *savecp;
+      EXPR_CONTEXT ec;
+      int peektok;
 
       while (legal_variable_char (c))
 	c = *cp++;
@@ -827,30 +910,47 @@ readtok ()
 #endif /* ARRAY_VARS */
 
       *cp = '\0';
-
       FREE (tokstr);
       tokstr = savestring (tp);
-
-#if defined (ARRAY_VARS)
-      value = (e == ']') ? get_array_value (tokstr, 0) : get_string_value (tokstr);
-#else
-      value = get_string_value (tokstr);
-#endif
-
-      tokval = (value && *value) ? subexpr (value) : 0;
-
-#if defined (ARRAY_VARS)
-      if (e == ']')
-	FREE (value);	/* get_array_value returns newly-allocated memory */
-#endif
-
       *cp = c;
+
+      SAVETOK (&ec);
+      tokstr = (char *)NULL;	/* keep it from being freed */
+      tp = savecp = cp;
+      noeval = 1;
+      readtok ();
+      peektok = curtok;
+      if (peektok == STR)	/* free new tokstr before old one is restored */
+	FREE (tokstr);
+      RESTORETOK (&ec);
+      cp = savecp;
+
+      /* The tests for PREINC and PREDEC aren't strictly correct, but they
+	 preserve old behavior if a construct like --x=9 is given. */
+      if (lasttok == PREINC || lasttok == PREDEC || peektok != EQ)
+	{
+#if defined (ARRAY_VARS)
+	  value = (e == ']') ? get_array_value (tokstr, 0) : get_string_value (tokstr);
+#else
+	  value = get_string_value (tokstr);
+#endif
+
+	  tokval = (value && *value) ? subexpr (value) : 0;
+
+#if defined (ARRAY_VARS)
+	  if (e == ']')
+	    FREE (value);	/* get_array_value returns newly-allocated memory */
+#endif
+	}
+      else
+	tokval = 0;
+
       lasttok = curtok;
       curtok = STR;
     }
-  else if (digit(c))
+  else if (DIGIT(c))
     {
-      while (digit (c) || isletter (c) || c == '#' || c == '@' || c == '_')
+      while (ISALNUM (c) || c == '#' || c == '@' || c == '_')
 	c = *cp++;
 
       c = *--cp;
@@ -899,8 +999,12 @@ readtok ()
       else if ((c == BOR) && (c1 == BOR))
 	c = LOR;
       else if ((c == '*') && (c1 == '*'))
-        c = POWER;
-      else if (c1 == EQ && member(c, "*/%+-&^|"))
+	c = POWER;
+      else if ((c == '-') && (c1 == '-') && legal_variable_starter ((unsigned char)*cp))
+	c = PREDEC;
+      else if ((c == '+') && (c1 == '+') && legal_variable_starter ((unsigned char)*cp))
+	c = PREINC;
+      else if (c1 == EQ && member (c, "*/%+-&^|"))
 	{
 	  assigntok = c;	/* a OP= b */
 	  c = OP_ASSIGN;
@@ -944,13 +1048,11 @@ strlong (num)
      char *num;
 {
   register char *s;
-  register int c;
+  register unsigned char c;
   int base, foundbase;
-  long val = 0L;
+  long val;
 
   s = num;
-  if (s == NULL || *s == '\0')
-    return 0L;
 
   base = 10;
   foundbase = 0;
@@ -958,8 +1060,8 @@ strlong (num)
     {
       s++;
 
-      if (s == NULL || *s == '\0')
-	return 0L;
+      if (*s == '\0')
+	return 0;
 
        /* Base 16? */
       if (*s == 'x' || *s == 'X')
@@ -972,7 +1074,7 @@ strlong (num)
       foundbase++;
     }
 
-  val = 0L;
+  val = 0;
   for (c = *s++; c; c = *s++)
     {
       if (c == '#')
@@ -980,26 +1082,25 @@ strlong (num)
 	  if (foundbase)
 	    evalerror ("bad number");
 
-	  base = (int)val;
-
 	  /* Illegal base specifications raise an evaluation error. */
-	  if (base < 2 || base > 64)
+	  if (val < 2 || val > 64)
 	    evalerror ("illegal arithmetic base");
 
-	  val = 0L;
+	  base = val;
+	  val = 0;
 	  foundbase++;
 	}
-      else if (isletter(c) || digit(c) || (c == '_') || (c == '@'))
+      else if (ISALNUM(c) || (c == '_') || (c == '@'))
 	{
-	  if (digit(c))
-	    c = digit_value(c);
+	  if (DIGIT(c))
+	    c = TODIGIT(c);
 	  else if (c >= 'a' && c <= 'z')
 	    c -= 'a' - 10;
 	  else if (c >= 'A' && c <= 'Z')
 	    c -= 'A' - ((base <= 36) ? 10 : 36);
-	  else if (c == '_')
-	    c = 62;
 	  else if (c == '@')
+	    c = 62;
+	  else if (c == '_')
 	    c = 63;
 
 	  if (c >= base)
@@ -1014,14 +1115,14 @@ strlong (num)
 }
 
 #if defined (EXPR_TEST)
-char *
+void *
 xmalloc (n)
      int n;
 {
   return (malloc (n));
 }
 
-char *
+void *
 xrealloc (s, n)
      char *s;
      int n;
@@ -1051,9 +1152,9 @@ main (argc, argv)
     {
       v = evalexp (argv[i], &expok);
       if (expok == 0)
-        fprintf (stderr, "%s: expression error\n", argv[i]);
+	fprintf (stderr, "%s: expression error\n", argv[i]);
       else
-        printf ("'%s' -> %ld\n", argv[i], v);
+	printf ("'%s' -> %ld\n", argv[i], v);
     }
   exit (0);
 }
@@ -1070,7 +1171,7 @@ builtin_error (format, arg1, arg2, arg3, arg4, arg5)
 
 char *
 itos (n)
-     int n;
+     long n;
 {
   return ("42");
 }
