@@ -22,7 +22,7 @@
 #include "config.h"
 
 #include <stdio.h>
-#include <ctype.h>
+#include "chartypes.h"
 #include "bashtypes.h"
 #ifndef _MINIX
 #  include <sys/file.h>
@@ -34,10 +34,6 @@
 #  include <unistd.h>
 #endif
 
-#if defined (HAVE_LIMITS_H)
-#  include <limits.h>
-#endif
-
 #include "bashansi.h"
 
 #include "memalloc.h"
@@ -46,12 +42,18 @@
 #include "hashlib.h"
 #include "pathexp.h"
 #include "hashcmd.h"
+#include "findcmd.h"	/* matching prototypes and declarations */
 
 extern int posixly_correct;
 
 /* Static functions defined and used in this file. */
-static char *find_user_command_internal (), *find_user_command_in_path ();
-static char *find_in_path_element (), *find_absolute_program ();
+static char *_find_user_command_internal __P((const char *, int));
+static char *find_user_command_internal __P((const char *, int));
+static char *find_user_command_in_path __P((const char *, char *, int));
+static char *find_in_path_element __P((const char *, char *, int, int, struct stat *));
+static char *find_absolute_program __P((const char *, int));
+
+static char *get_next_path_element __P((char *, int *));
 
 /* The file name which we would try to execute, except that it isn't
    possible to execute it.  This is the first file that matches the
@@ -81,7 +83,7 @@ int dot_found_in_search = 0;
    Zero is returned if the file is not found. */
 int
 file_status (name)
-     char *name;
+     const char *name;
 {
   struct stat finfo;
 
@@ -145,7 +147,7 @@ file_status (name)
    what an executable file is. */
 int
 executable_file (file)
-     char *file;
+     const char *file;
 {
   int s;
 
@@ -155,11 +157,21 @@ executable_file (file)
 
 int
 is_directory (file)
-     char *file;
+     const char *file;
 {
   return (file_status (file) & FS_DIRECTORY);
 }
 
+int
+executable_or_directory (file)
+     const char *file;
+{
+  int s;
+
+  s = file_status (file);
+  return ((s & FS_EXECABLE) || (s & FS_DIRECTORY));
+}
+
 /* Locate the executable file referenced by NAME, searching along
    the contents of the shell PATH variable.  Return a new string
    which is the full pathname to the file, or NULL if the file
@@ -167,7 +179,7 @@ is_directory (file)
    and that is the only match, then return that. */
 char *
 find_user_command (name)
-     char *name;
+     const char *name;
 {
   return (find_user_command_internal (name, FS_EXEC_PREFERRED|FS_NODIRS));
 }
@@ -178,20 +190,20 @@ find_user_command (name)
    returns the first file found. */
 char *
 find_path_file (name)
-     char *name;
+     const char *name;
 {
   return (find_user_command_internal (name, FS_EXISTS));
 }
 
 static char *
 _find_user_command_internal (name, flags)
-     char *name;
+     const char *name;
      int flags;
 {
   char *path_list, *cmd;
   SHELL_VAR *var;
 
-  /* Search for the value of PATH in both the temporary environment, and
+  /* Search for the value of PATH in both the temporary environments and
      in the regular list of variables. */
   if (var = find_variable_internal ("PATH", 1))	/* XXX could be array? */
     path_list = value_cell (var);
@@ -203,21 +215,18 @@ _find_user_command_internal (name, flags)
 
   cmd = find_user_command_in_path (name, path_list, flags);
 
-  if (var && tempvar_p (var))
-    dispose_variable (var);
-
   return (cmd);
 }
 
 static char *
 find_user_command_internal (name, flags)
-     char *name;
+     const char *name;
      int flags;
 {
 #ifdef __WIN32__
   char *res, *dotexe;
 
-  dotexe = xmalloc (strlen (name) + 5);
+  dotexe = (char *)xmalloc (strlen (name) + 5);
   strcpy (dotexe, name);
   strcat (dotexe, ".exe");
   res = _find_user_command_internal (dotexe, flags);
@@ -260,7 +269,7 @@ get_next_path_element (path_list, path_index_pointer)
    in $PATH.  Returns a newly-allocated string. */
 char *
 search_for_command (pathname)
-     char *pathname;
+     const char *pathname;
 {
   char *hashed_file, *command;
   int temp_path, st;
@@ -270,14 +279,16 @@ search_for_command (pathname)
 
   /* If PATH is in the temporary environment for this command, don't use the
      hash table to search for the full pathname. */
-  path = find_tempenv_variable ("PATH");
-  temp_path = path != 0;
+  path = find_variable_internal ("PATH", 1);
+  temp_path = path && tempvar_p (path);
+  if (temp_path == 0 && path)
+    path = (SHELL_VAR *)NULL;
 
   /* Don't waste time trying to find hashed data for a pathname
      that is already completely specified or if we're using a command-
      specific value for PATH. */
   if (path == 0 && absolute_program (pathname) == 0)
-    hashed_file = find_hashed_filename (pathname);
+    hashed_file = phash_search (pathname);
 
   /* If a command found in the hash table no longer exists, we need to
      look for it in $PATH.  Thank you Posix.2.  This forces us to stat
@@ -288,7 +299,7 @@ search_for_command (pathname)
       st = file_status (hashed_file);
       if ((st ^ (FS_EXISTS | FS_EXECABLE)) != 0)
 	{
-	  remove_hashed_filename (pathname);
+	  phash_remove (pathname);
 	  free (hashed_file);
 	  hashed_file = (char *)NULL;
 	}
@@ -305,23 +316,21 @@ search_for_command (pathname)
       /* If $PATH is in the temporary environment, we've already retrieved
 	 it, so don't bother trying again. */
       if (temp_path)
-        {
+	{
 	  command = find_user_command_in_path (pathname, value_cell (path),
 					       FS_EXEC_PREFERRED|FS_NODIRS);
-	  if (tempvar_p (path))
-	    dispose_variable (path);
-        }
+	}
       else
 	command = find_user_command (pathname);
       if (command && hashing_enabled && temp_path == 0)
-	remember_filename (pathname, command, dot_found_in_search, 1);
+	phash_insert ((char *)pathname, command, dot_found_in_search, 1);	/* XXX fix const later */
     }
   return (command);
 }
 
 char *
 user_command_matches (name, flags, state)
-     char *name;
+     const char *name;
      int flags, state;
 {
   register int i;
@@ -338,7 +347,7 @@ user_command_matches (name, flags, state)
       if (match_list == 0)
 	{
 	  match_list_size = 5;
-	  match_list = alloc_array (match_list_size);
+	  match_list = strvec_create (match_list_size);
 	}
 
       /* Clear out the old match list. */
@@ -381,7 +390,7 @@ user_command_matches (name, flags, state)
 	  if (match_index + 1 == match_list_size)
 	    {
 	      match_list_size += 10;
-	      match_list = (char **)xrealloc (match_list, (match_list_size + 1) * sizeof (char *));
+	      match_list = strvec_resize (match_list, (match_list_size + 1));
 	    }
 
 	  match_list[match_index++] = match;
@@ -404,7 +413,7 @@ user_command_matches (name, flags, state)
 
 static char *
 find_absolute_program (name, flags)
-     char *name;
+     const char *name;
      int flags;
 {
   int st;
@@ -421,19 +430,20 @@ find_absolute_program (name, flags)
   if ((flags & FS_EXISTS) || ((flags & FS_EXEC_ONLY) && (st & FS_EXECABLE)))
     return (savestring (name));
 
-  return ((char *)NULL);
+  return (NULL);
 }
 
 static char *
 find_in_path_element (name, path, flags, name_len, dotinfop)
-     char *name, *path;
+     const char *name;
+     char *path;
      int flags, name_len;
      struct stat *dotinfop;
 {
   int status;
   char *full_path, *xpath;
 
-  xpath = (*path == '~') ? bash_tilde_expand (path) : path;
+  xpath = (*path == '~') ? bash_tilde_expand (path, 0) : path;
 
   /* Remember the location of "." in the path, in all its forms
      (as long as they begin with a `.', e.g. `./.') */
@@ -499,7 +509,7 @@ find_in_path_element (name, path, flags, name_len, dotinfop)
 */
 static char *
 find_user_command_in_path (name, path_list, flags)
-     char *name;
+     const char *name;
      char *path_list;
      int flags;
 {
@@ -557,6 +567,14 @@ find_user_command_in_path (name, path_list, flags)
   /* We didn't find exactly what the user was looking for.  Return
      the contents of FILE_TO_LOSE_ON which is NULL when the search
      required an executable, or non-NULL if a file was found and the
-     search would accept a non-executable as a last resort. */
+     search would accept a non-executable as a last resort.  If the
+     caller specified FS_NODIRS, and file_to_lose_on is a directory,
+     return NULL. */
+  if (file_to_lose_on && (flags & FS_NODIRS) && is_directory (file_to_lose_on))
+    {
+      free (file_to_lose_on);
+      file_to_lose_on = (char *)NULL;
+    }
+
   return (file_to_lose_on);
 }
