@@ -27,10 +27,13 @@
 #  define _FUNCTION_DEF
 typedef int Function ();
 typedef void VFunction ();
-typedef char *CPFunction ();
-typedef char **CPPFunction ();
+typedef char *CPFunction ();		/* no longer used */
+typedef char **CPPFunction ();		/* no longer used */
 #endif /* _FUNCTION_DEF */
 
+typedef int sh_cget_func_t __P((void));		/* sh_ivoidfunc_t */
+typedef int sh_cunget_func_t __P((int));	/* sh_intfunc_t */
+
 enum stream_type {st_none, st_stdin, st_stream, st_string, st_bstream};
 
 #if defined (BUFFERED_INPUT)
@@ -40,20 +43,21 @@ enum stream_type {st_none, st_stdin, st_stream, st_string, st_bstream};
 #undef B_ERROR		/* There are some systems with this define */
 #undef B_UNBUFF
 
-#define B_EOF		0x1
-#define B_ERROR		0x2
-#define B_UNBUFF	0x4
+#define B_EOF		0x01
+#define B_ERROR		0x02
+#define B_UNBUFF	0x04
+#define B_WASBASHINPUT	0x08
 
 /* A buffered stream.  Like a FILE *, but with our own buffering and
    synchronization.  Look in input.c for the implementation. */
 typedef struct BSTREAM
 {
-  int	b_fd;
+  int	 b_fd;
   char	*b_buffer;		/* The buffer that holds characters read. */
   size_t b_size;		/* How big the buffer is. */
-  int	b_used;			/* How much of the buffer we're using, */
-  int	b_flag;			/* Flag values. */
-  int	b_inputp;		/* The input pointer, index into b_buffer. */
+  size_t b_used;		/* How much of the buffer we're using, */
+  int	 b_flag;		/* Flag values. */
+  size_t b_inputp;		/* The input pointer, index into b_buffer. */
 } BUFFERED_STREAM;
 
 #if 0
@@ -76,18 +80,18 @@ typedef struct {
   enum stream_type type;
   char *name;
   INPUT_STREAM location;
-  Function *getter;
-  Function *ungetter;
+  sh_cget_func_t *getter;
+  sh_cunget_func_t *ungetter;
 } BASH_INPUT;
 
 extern BASH_INPUT bash_input;
 
 /* Functions from parse.y. */
 extern void initialize_bash_input __P((void));
-extern void init_yy_io __P((Function *, Function *, enum stream_type, char *, INPUT_STREAM));
+extern void init_yy_io __P((sh_cget_func_t *, sh_cunget_func_t *, enum stream_type, const char *, INPUT_STREAM));
 extern void with_input_from_stdin __P((void));
-extern void with_input_from_string __P((char *, char *));
-extern void with_input_from_stream __P((FILE *, char *));
+extern void with_input_from_string __P((char *, const char *));
+extern void with_input_from_stream __P((FILE *, const char *));
 extern void push_stream __P((int));
 extern void pop_stream __P((void));
 extern int stream_on_stack __P((enum stream_type));
@@ -101,11 +105,14 @@ extern int *save_token_state __P((void));
 extern void restore_token_state __P((int *));
 
 /* Functions from input.c */
-extern int getc_with_restart ();
-extern int ungetc_with_restart ();
+extern int getc_with_restart __P((FILE *));
+extern int ungetc_with_restart __P((int, FILE *));
 
 #if defined (BUFFERED_INPUT)
 /* Functions from input.c. */
+extern int fd_is_bash_input __P((int));
+extern int set_bash_input_fd __P((int));
+extern int save_bash_input __P((int, int));
 extern int check_bash_input __P((int));
 extern int duplicate_buffered_stream __P((int, int));
 extern BUFFERED_STREAM *fd_to_buffered_stream __P((int));
