@@ -16,7 +16,7 @@
 
    You should have received a copy of the GNU General Public License along
    with Bash; see the file COPYING.  If not, write to the Free Software
-   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */
+   Foundation, 59 Temple Place, Suite 330, Boston, MA 02111 USA. */
 
 #if !defined (_JOBS_H_)
 #  define _JOBS_H_
@@ -42,6 +42,11 @@
    of processes that are piped together.  The first process encountered is
    the group leader. */
 
+/* Values for the `running' field of a struct process. */
+#define PS_DONE		0
+#define PS_RUNNING	1
+#define PS_STOPPED	2
+
 /* Each child of the shell is remembered in a STRUCT PROCESS.  A chain of
    such structures is a pipeline.  The chain is circular. */
 typedef struct process {
@@ -78,7 +83,7 @@ typedef struct job {
   int flags;	   /* Flags word: J_NOTIFIED, J_FOREGROUND, or J_JOBCONTROL. */
 #if defined (JOB_CONTROL)
   COMMAND *deferred;	/* Commands that will execute when this job is done. */
-  VFunction *j_cleanup; /* Cleanup function to call when job marked JDEAD */
+  sh_vptrfunc_t *j_cleanup; /* Cleanup function to call when job marked JDEAD */
   PTR_T cleanarg;	/* Argument passed to (*j_cleanup)() */
 #endif /* JOB_CONTROL */
 } JOB;
@@ -95,7 +100,7 @@ extern pid_t fork (), getpid (), getpgrp ();
 #endif /* !HAVE_UNISTD_H */
 
 /* Stuff from the jobs.c file. */
-extern pid_t  original_pgrp, shell_pgrp, pipeline_pgrp;
+extern pid_t original_pgrp, shell_pgrp, pipeline_pgrp;
 extern pid_t last_made_pid, last_asynchronous_pid;
 extern int current_job, previous_job;
 extern int asynchronous_notification;
@@ -115,6 +120,8 @@ extern void nohup_job __P((int));
 extern void delete_all_jobs __P((int));
 extern void nohup_all_jobs __P((int));
 
+extern int count_all_jobs __P((void));
+
 extern void terminate_current_pipeline __P((void));
 extern void terminate_stopped_jobs __P((void));
 extern void hangup_all_jobs __P((void));
@@ -134,6 +141,7 @@ extern void list_stopped_jobs __P((int));
 extern void list_running_jobs __P((int));
 
 extern pid_t make_child __P((char *, int));
+
 extern int get_tty_state __P((void));
 extern int set_tty_state __P((void));
 
@@ -148,7 +156,7 @@ extern int start_job __P((int, int));
 extern int kill_pid __P((pid_t, int, int));
 extern int initialize_job_control __P((int));
 extern void initialize_job_signals __P((void));
-extern int give_terminal_to __P((pid_t));
+extern int give_terminal_to __P((pid_t, int));
 
 extern void set_sigwinch_handler __P((void));
 extern void unset_sigwinch_handler __P((void));
