@@ -25,10 +25,10 @@
 
 #include "rlstdc.h"
 
-extern char *single_quote __P((char *));
-extern void set_lines_and_columns __P((int, int));
-extern char *get_env_value __P((char *));
-extern char *get_home_dir __P((void));
-extern int unset_nodelay_mode __P((int));
+extern char *sh_single_quote __P((char *));
+extern void sh_set_lines_and_columns __P((int, int));
+extern char *sh_get_env_value __P((const char *));
+extern char *sh_get_home_dir __P((void));
+extern int sh_unset_nodelay_mode __P((int));
 
 #endif /* _RL_SHELL_H_ */
