@@ -18,61 +18,56 @@
    with Bash; see the file COPYING.  If not, write to the Free Software
    Foundation, 59 Temple Place, Suite 330, Boston, MA 02111 USA. */
 
-#include <config.h>
+#ifdef HAVE_CONFIG_H
+#  include <config.h>
+#endif
 
 #if defined (HAVE_UNISTD_H)
 #  include <unistd.h>
 #endif
 
-#include "bashansi.h"
+#include <bashansi.h>
+#include <chartypes.h>
 #include "shell.h"
 
-/* Number of characters that can appear in a string representation
-   of an integer.  32 is larger than the string rep of 2^^31 - 1. */
-#define MAX_INT_LEN 32
-
-/* Integer to string conversion.  The caller passes the buffer and
-   the size.  This should check for buffer underflow, but currently
-   does not. */
 char *
 inttostr (i, buf, len)
-     int i;
+     long i;
      char *buf;
-     int len;
+     size_t len;
 {
-  char *p;
-  int negative = 0;
-  unsigned int ui;
-
-  if (i < 0)
-    {
-      negative++;
-      i = -i;
-    }
-
-  ui = (unsigned int) i;
-
-  p = buf + len - 2;
-  p[1] = '\0';
-
-  do
-    *p-- = (ui % 10) + '0';
-  while (ui /= 10);
-
-  if (negative)
-    *p-- = '-';
-
-  return (p + 1);
+  return (fmtulong (i, 10, buf, len, 0));
 }
 
 /* Integer to string conversion.  This conses the string; the
    caller should free it. */
 char *
 itos (i)
-     int i;
+     long i;
 {
-  char *p, lbuf[MAX_INT_LEN];
+  char *p, lbuf[INT_STRLEN_BOUND(long) + 1];
 
-  p = inttostr (i, lbuf, sizeof(lbuf));
+  p = fmtulong (i, 10, lbuf, sizeof(lbuf), 0);
+  return (savestring (p));
+}
+
+char *
+uinttostr (i, buf, len)
+     unsigned long i;
+     char *buf;
+     size_t len;
+{
+  return (fmtulong (i, 10, buf, len, FL_UNSIGNED));
+}
+
+/* Integer to string conversion.  This conses the string; the
+   caller should free it. */
+char *
+uitos (i)
+     unsigned long i;
+{
+  char *p, lbuf[INT_STRLEN_BOUND(long) + 1];
+
+  p = fmtulong (i, 10, lbuf, sizeof(lbuf), FL_UNSIGNED);
   return (savestring (p));
 }
