@@ -27,11 +27,8 @@
 
 #include <stdio.h>
 
-#if !defined(slashify_in_quotes)
-#  define slashify_in_quotes "\\`$\"\n"
-#endif
-
-extern char *xmalloc ();
+#include "syntax.h"
+#include <xmalloc.h>
 
 /* **************************************************************** */
 /*								    */
@@ -42,13 +39,13 @@ extern char *xmalloc ();
 /* Return a new string which is the single-quoted version of STRING.
    Used by alias and trap, among others. */
 char *
-single_quote (string)
+sh_single_quote (string)
      char *string;
 {
   register int c;
   char *result, *r, *s;
 
-  result = xmalloc (3 + (4 * strlen (string)));
+  result = (char *)xmalloc (3 + (4 * strlen (string)));
   r = result;
   *r++ = '\'';
 
@@ -72,30 +69,22 @@ single_quote (string)
 
 /* Quote STRING using double quotes.  Return a new string. */
 char *
-double_quote (string)
+sh_double_quote (string)
      char *string;
 {
-  register int c;
+  register unsigned char c;
   char *result, *r, *s;
 
-  result = xmalloc (3 + (2 * strlen (string)));
+  result = (char *)xmalloc (3 + (2 * strlen (string)));
   r = result;
   *r++ = '"';
 
   for (s = string; s && (c = *s); s++)
     {
-      switch (c)
-        {
-	case '"':
-	case '$':
-	case '`':
-	case '\\':
-	case '\n':		/* XXX */
-	  *r++ = '\\';
-	default:
-	  *r++ = c;
-	  break;
-        }
+      if (sh_syntaxtab[c] & CBSDQUOTE)
+	*r++ = '\\';
+
+      *r++ = c;
     }
 
   *r++ = '"';
@@ -107,13 +96,13 @@ double_quote (string)
 /* Remove backslashes that are quoting characters that are special between
    double quotes.  Return a new string. */
 char *
-un_double_quote (string)
+sh_un_double_quote (string)
      char *string;
 {
   register int c, pass_next;
   char *result, *r, *s;
 
-  r = result = xmalloc (strlen (string) + 1);
+  r = result = (char *)xmalloc (strlen (string) + 1);
 
   for (pass_next = 0, s = string; s && (c = *s); s++)
     {
@@ -123,7 +112,7 @@ un_double_quote (string)
 	  pass_next = 0;
 	  continue;
 	}
-      if (c == '\\' && strchr (slashify_in_quotes, s[1]))
+      if (c == '\\' && (sh_syntaxtab[(unsigned char) s[1]] & CBSDQUOTE))
 	{
 	  pass_next = 1;
 	  continue;
@@ -138,13 +127,13 @@ un_double_quote (string)
 /* Quote special characters in STRING using backslashes.  Return a new
    string. */
 char *
-backslash_quote (string)
+sh_backslash_quote (string)
      char *string;
 {
   int c;
   char *result, *r, *s;
 
-  result = xmalloc (2 * strlen (string) + 1);
+  result = (char *)xmalloc (2 * strlen (string) + 1);
 
   for (r = result, s = string; s && (c = *s); s++)
     {
@@ -182,8 +171,33 @@ backslash_quote (string)
   return (result);
 }
 
+#if defined (PROMPT_STRING_DECODE)
+/* Quote characters that get special treatment when in double quotes in STRING
+   using backslashes.  Return a new string. */
+char *
+sh_backslash_quote_for_double_quotes (string)
+     char *string;
+{
+  unsigned char c;
+  char *result, *r, *s;
+
+  result = (char *)xmalloc (2 * strlen (string) + 1);
+
+  for (r = result, s = string; s && (c = *s); s++)
+    {
+      if (sh_syntaxtab[c] & CBSDQUOTE)
+	*r++ = '\\';
+
+      *r++ = c;
+    }
+
+  *r = '\0';
+  return (result);
+}
+#endif /* PROMPT_STRING_DECODE */
+
 int
-contains_shell_metas (string)
+sh_contains_shell_metas (string)
      char *string;
 {
   char *s;
@@ -204,6 +218,7 @@ contains_shell_metas (string)
 	case '~':				/* tilde expansion */
 	  if (s == string || s[-1] == '=' || s[-1] == ':')
 	    return (1);
+	  break;
 	case '#':
 	  if (s == string)			/* comment char */
 	    return (1);
