@@ -29,7 +29,7 @@
 #include "bashintl.h"
 #include "bashansi.h"
 #include <stdio.h>
-#include <ctype.h>
+#include "chartypes.h"
 
 #include "shell.h"
 
@@ -138,7 +138,7 @@ set_locale_var (var, value)
 	lc_all = savestring (default_locale);
       else
 	{
-	  lc_all = xmalloc (1);
+	  lc_all = (char *)xmalloc (1);
 	  lc_all[0] = '\0';
 	}
 #if defined (HAVE_SETLOCALE)
@@ -168,7 +168,7 @@ set_locale_var (var, value)
 	return (setlocale (LC_MESSAGES, value ? value : "") != 0);
 #  endif /* LC_MESSAGES */
     }
-  else if (var[3] = 'N' && var[4] == 'U')	/* LC_NUMERIC */
+  else if (var[3] == 'N' && var[4] == 'U')	/* LC_NUMERIC */
     {
 #  if defined (LC_NUMERIC)
       if (lc_all == 0 || *lc_all == '\0')
@@ -234,12 +234,10 @@ localetrans (string, len, lenp)
   if (string == 0 || *string == 0)
     {
       if (lenp)
-        *lenp = 0;
+	*lenp = 0;
       return ((char *)NULL);
     }
 
-  t = xmalloc (len + 1);
-
   locale = get_locale_var ("LC_MESSAGES");
 
   /* If we don't have setlocale() or the current locale is `C' or `POSIX',
@@ -250,6 +248,7 @@ localetrans (string, len, lenp)
       (locale[0] == 'C' && locale[1] == '\0') || STREQ (locale, "POSIX"))
 #endif
     {
+      t = (char *)xmalloc (len + 1);
       strcpy (t, string);
       if (lenp)
 	*lenp = len;
@@ -261,15 +260,15 @@ localetrans (string, len, lenp)
   translated = gettext (string);
   if (translated == string)	/* gettext returns its argument if untranslatable */
     {
+      t = (char *)xmalloc (len + 1);
       strcpy (t, string);
       if (lenp)
 	*lenp = len;
     }
   else
     {
-      free (t);
       tlen = strlen (translated);
-      t = xmalloc (tlen + 1);
+      t = (char *)xmalloc (tlen + 1);
       strcpy (t, translated);
       if (lenp)
 	*lenp = tlen;
