@@ -33,10 +33,11 @@
 #include "pathexp.h"
 #include "flags.h"
 
-#include <glob/fnmatch.h>
+#include <glob/strmatch.h>
 
 #if defined (USE_POSIX_GLOB_LIBRARY)
 #  include <glob.h>
+typedef int posix_glob_errfunc_t __P((const char *, int));
 #else
 #  include <glob/glob.h>
 #endif
@@ -101,13 +102,13 @@ unquoted_glob_pattern_p (string)
    to match a filename should be performed. */
 char *
 quote_string_for_globbing (pathname, qflags)
-     char *pathname;
+     const char *pathname;
      int qflags;
 {
   char *temp;
   register int i, j;
 
-  temp = xmalloc (strlen (pathname) + 1);
+  temp = (char *)xmalloc (strlen (pathname) + 1);
 
   if ((qflags & QGLOB_CVTNULL) && QUOTED_NULL (pathname))
     {
@@ -118,13 +119,13 @@ quote_string_for_globbing (pathname, qflags)
   for (i = j = 0; pathname[i]; i++)
     {
       if (pathname[i] == CTLESC)
-        {
-          if ((qflags & QGLOB_FILENAME) && pathname[i+1] == '/')
-            continue;
+	{
+	  if ((qflags & QGLOB_FILENAME) && pathname[i+1] == '/')
+	    continue;
 	  temp[j++] = '\\';
-        }
+	}
       else
-        temp[j++] = pathname[i];
+	temp[j++] = pathname[i];
     }
   temp[j] = '\0';
 
@@ -137,25 +138,25 @@ quote_globbing_chars (string)
 {
   char *temp, *s, *t;
 
-  temp = xmalloc (strlen (string) * 2 + 1);
+  temp = (char *)xmalloc (strlen (string) * 2 + 1);
   for (t = temp, s = string; *s; )
     {
       switch (*s)
-        {
-        case '*':
-        case '[':
-        case ']':
-        case '?':
-        case '\\':
-          *t++ = '\\';
-          break;
-        case '+':
-        case '@':
-        case '!':
+	{
+	case '*':
+	case '[':
+	case ']':
+	case '?':
+	case '\\':
+	  *t++ = '\\';
+	  break;
+	case '+':
+	case '@':
+	case '!':
 	  if (s[1] == '(')	/*(*/
 	    *t++ = '\\';
 	  break;
-        }
+	}
       *t++ = *s++;
     }
   *t = '\0';
@@ -165,11 +166,11 @@ quote_globbing_chars (string)
 /* Call the glob library to do globbing on PATHNAME. */
 char **
 shell_glob_filename (pathname)
-     char *pathname;
+     const char *pathname;
 {
 #if defined (USE_POSIX_GLOB_LIBRARY)
   register int i;
-  char *temp, **return_value;
+  char *temp, **results;
   glob_t filenames;
   int glob_flags;
 
@@ -185,11 +186,11 @@ shell_glob_filename (pathname)
 
   glob_flags |= (GLOB_ERR | GLOB_DOOFFS);
 
-  i = glob (temp, glob_flags, (Function *)NULL, &filenames);
+  i = glob (temp, glob_flags, (posix_glob_errfunc_t *)NULL, &filenames);
 
   free (temp);
 
-  if (i == GLOB_NOSPACE || i == GLOB_ABEND)
+  if (i == GLOB_NOSPACE || i == GLOB_ABORTED)
     return ((char **)NULL);
   else if (i == GLOB_NOMATCH)
     filenames.gl_pathv = (char **)NULL;
@@ -249,7 +250,7 @@ static struct ignorevar globignore =
   (struct ign *)0,
   0,
   (char *)0,
-  (Function *)0,
+  (sh_iv_item_func_t *)0,
 };
 
 /* Set up to ignore some glob matches because the value of GLOBIGNORE
@@ -279,7 +280,7 @@ should_ignore_glob_matches ()
 /* Return 0 if NAME matches a pattern in the globignore.ignores list. */
 static int
 glob_name_is_acceptable (name)
-     char *name;
+     const char *name;
 {
   struct ign *p;
   int flags;
@@ -291,8 +292,8 @@ glob_name_is_acceptable (name)
   flags = FNM_PATHNAME | FNMATCH_EXTFLAG;
   for (p = globignore.ignores; p->val; p++)
     {
-      if (fnmatch (p->val, name, flags) != FNM_NOMATCH)
-        return (0);
+      if (strmatch (p->val, (char *)name, flags) != FNM_NOMATCH)
+	return (0);
     }
   return (1);
 }
@@ -306,7 +307,7 @@ glob_name_is_acceptable (name)
 static void
 ignore_globbed_names (names, name_func)
      char **names;
-     Function *name_func;
+     sh_ignore_func_t *name_func;
 {
   char **newnames;
   int n, i;
@@ -318,7 +319,7 @@ ignore_globbed_names (names, name_func)
   for (n = i = 0; names[i]; i++)
     {
       if ((*name_func) (names[i]))
-        newnames[n++] = names[i];
+	newnames[n++] = names[i];
       else
 	free (names[i]);
     }
@@ -400,7 +401,7 @@ setup_ignore_patterns (ivp)
       ivp->ignores[numitems].len = strlen (colon_bit);
       ivp->ignores[numitems].flags = 0;
       if (ivp->item_func)
-        (*ivp->item_func) (&ivp->ignores[numitems]);
+	(*ivp->item_func) (&ivp->ignores[numitems]);
       numitems++;
     }
   ivp->ignores[numitems].val = (char *)NULL;
