@@ -56,12 +56,12 @@ extern int unquoted_glob_pattern_p __P((char *));
    pattern while executing a case statement), flags should include
    QGLOB_CVTNULL.  If flags includes QGLOB_FILENAME, appropriate quoting
    to match a filename should be performed. */
-extern char *quote_string_for_globbing __P((char *, int));
+extern char *quote_string_for_globbing __P((const char *, int));
 
 extern char *quote_globbing_chars __P((char *));
 
 /* Call the glob library to do globbing on PATHNAME. */
-extern char **shell_glob_filename __P((char *));
+extern char **shell_glob_filename __P((const char *));
 
 /* Filename completion ignore.  Used to the "fignore" facility of
    tcsh and GLOBIGNORE (like ksh-93 FIGNORE).
