@@ -47,6 +47,7 @@
 #include "pcomplete.h"
 #include "alias.h"
 #include "bashline.h"
+#include "execute_cmd.h"
 #include "pathexp.h"
 
 #if defined (JOB_CONTROL)
@@ -61,7 +62,7 @@
 #include "builtins/common.h"
 
 #include <glob/glob.h>
-#include <glob/fnmatch.h>
+#include <glob/strmatch.h>
 
 #include <readline/rlconf.h>
 #include <readline/readline.h>
@@ -78,28 +79,58 @@ typedef SHELL_VAR **SVFUNC ();
 extern char *strpbrk __P((char *, char *));
 #endif
 
-extern int rl_filename_completion_desired;
 extern int array_needs_making;
 extern STRING_INT_ALIST word_token_alist[];
 extern char *signal_names[];
 
-static int it_init_aliases ();
-static int it_init_arrayvars ();
-static int it_init_bindings ();
-static int it_init_builtins ();
-static int it_init_disabled ();
-static int it_init_enabled ();
-static int it_init_exported ();
-static int it_init_functions ();
-static int it_init_hostnames ();
-static int it_init_jobs ();
-static int it_init_running ();
-static int it_init_stopped ();
-static int it_init_keywords ();
-static int it_init_signals ();
-static int it_init_variables ();
-static int it_init_setopts ();
-static int it_init_shopts ();
+#if defined(PREFER_STDARG)
+static void debug_printf (const char *, ...)  __attribute__((__format__ (printf, 1, 2)));
+#endif
+
+static int it_init_joblist __P((ITEMLIST *, int));
+
+static int it_init_aliases __P((ITEMLIST *));
+static int it_init_arrayvars __P((ITEMLIST *));
+static int it_init_bindings __P((ITEMLIST *));
+static int it_init_builtins __P((ITEMLIST *));
+static int it_init_disabled __P((ITEMLIST *));
+static int it_init_enabled __P((ITEMLIST *));
+static int it_init_exported __P((ITEMLIST *));
+static int it_init_functions __P((ITEMLIST *));
+static int it_init_hostnames __P((ITEMLIST *));
+static int it_init_jobs __P((ITEMLIST *));
+static int it_init_running __P((ITEMLIST *));
+static int it_init_stopped __P((ITEMLIST *));
+static int it_init_keywords __P((ITEMLIST *));
+static int it_init_signals __P((ITEMLIST *));
+static int it_init_variables __P((ITEMLIST *));
+static int it_init_setopts __P((ITEMLIST *));
+static int it_init_shopts __P((ITEMLIST *));
+
+static int shouldexp_filterpat __P((char *));
+static char *preproc_filterpat __P((char *, char *));
+
+static void init_itemlist_from_varlist __P((ITEMLIST *, SVFUNC *));
+
+static STRINGLIST *gen_matches_from_itemlist __P((ITEMLIST *, const char *));
+static STRINGLIST *gen_action_completions __P((COMPSPEC *, const char *));
+static STRINGLIST *gen_globpat_matches __P((COMPSPEC *, const char *));
+static STRINGLIST *gen_wordlist_matches __P((COMPSPEC *, const char *));
+static STRINGLIST *gen_shell_function_matches __P((COMPSPEC *, const char *,
+						   char *, int, WORD_LIST *,
+						   int, int));
+static STRINGLIST *gen_command_matches __P((COMPSPEC *, const char *, char *,
+					    int, WORD_LIST *, int, int));
+
+static char *pcomp_filename_completion_function __P((const char *, int));
+
+#if defined (ARRAY_VARS)
+static SHELL_VAR *bind_comp_words __P((WORD_LIST *));
+#endif
+static void bind_compfunc_variables __P((char *, int, WORD_LIST *, int, int));
+static void unbind_compfunc_variables __P((int));
+static WORD_LIST *build_arg_list __P((char *, const char *, WORD_LIST *, int));
+static WORD_LIST *command_line_to_word_list __P((char *, int, int, int *, int *));
 
 static int progcomp_debug = 0;
 
@@ -118,7 +149,8 @@ ITEMLIST it_exports  = { LIST_DYNAMIC, it_init_exported, (STRINGLIST *)0 };
 ITEMLIST it_files = { LIST_DYNAMIC };	/* unused */
 ITEMLIST it_functions  = { 0, it_init_functions, (STRINGLIST *)0 };
 ITEMLIST it_hostnames  = { LIST_DYNAMIC, it_init_hostnames, (STRINGLIST *)0 };
-ITEMLIST it_jobs = { LIST_DYNAMIC, it_init_jobs, (STRINGLIST *)0 };;
+ITEMLIST it_groups = { LIST_DYNAMIC };	/* unused */
+ITEMLIST it_jobs = { LIST_DYNAMIC, it_init_jobs, (STRINGLIST *)0 };
 ITEMLIST it_keywords = { 0, it_init_keywords, (STRINGLIST *)0 };
 ITEMLIST it_running = { LIST_DYNAMIC, it_init_running, (STRINGLIST *)0 };
 ITEMLIST it_setopts = { 0, it_init_setopts, (STRINGLIST *)0 };
@@ -209,132 +241,6 @@ clean_itemlist (itp)
   itp->flags &= ~(LIST_DONTFREE|LIST_DONTFREEMEMBERS|LIST_INITIALIZED|LIST_DIRTY);
 }
 
-/* Functions to manage the string lists -- lists of matches.  These should
-   really be moved to a separate file. */
-
-/* Allocate a new STRINGLIST, with room for N strings. */
-STRINGLIST *
-alloc_stringlist (n)
-     int n;
-{
-  STRINGLIST *ret;
-  register int i;
-
-  ret = (STRINGLIST *)xmalloc (sizeof (STRINGLIST));
-  if (n)
-    {
-      ret->list = alloc_array (n+1);
-      ret->list_size = n;
-      for (i = 0; i < n; i++)
-	ret->list[i] = (char *)NULL;
-    }
-  else
-    {
-      ret->list = (char **)NULL;
-      ret->list_size = 0;
-    }
-  ret->list_len = 0;
-  return ret;
-}
-
-STRINGLIST *
-realloc_stringlist (sl, n)
-     STRINGLIST *sl;
-     int n;
-{
-  register int i;
-
-  if (n > sl->list_size)
-    {
-      sl->list = (char **)xrealloc (sl->list, (n+1) * sizeof (char *));
-      for (i = sl->list_size; i <= n; i++)
-	sl->list[i] = (char *)NULL;
-      sl->list_size = n;
-    }
-  return sl;
-}
-  
-void
-free_stringlist (sl)
-     STRINGLIST *sl;
-{
-  if (sl == 0)
-    return;
-  if (sl->list)
-    free_array (sl->list);
-  free (sl);
-}
-
-STRINGLIST *
-copy_stringlist (sl)
-     STRINGLIST *sl;
-{
-  STRINGLIST *new;
-  register int i;
-
-  new = alloc_stringlist (sl->list_size);
-  /* I'd like to use copy_array, but that doesn't copy everything. */
-  if (sl->list)
-    {
-      for (i = 0; i < sl->list_size; i++)
-	new->list[i] = STRDUP (sl->list[i]);
-    }
-  new->list_size = sl->list_size;
-  new->list_len = sl->list_len;
-  /* just being careful */
-  if (new->list)
-    new->list[new->list_len] = (char *)NULL;
-  return new;
-}
-
-/* Return a new STRINGLIST with everything from M1 and M2. */
-
-STRINGLIST *
-merge_stringlists (m1, m2)
-     STRINGLIST *m1, *m2;
-{
-  STRINGLIST *sl;
-  int i, n, l1, l2;
-
-  l1 = m1 ? m1->list_len : 0;
-  l2 = m2 ? m2->list_len : 0;
-
-  sl = alloc_stringlist (l1 + l2 + 1);
-  for (i = n = 0; i < l1; i++, n++)
-    sl->list[n] = STRDUP (m1->list[i]);
-  for (i = 0; i < l2; i++, n++)
-    sl->list[n] = STRDUP (m2->list[i]);
-  sl->list_len = n;
-  sl->list[n] = (char *)NULL;
-}
-
-/* Make STRINGLIST M1 contain everything in M1 and M2. */
-STRINGLIST *
-append_stringlist (m1, m2)
-     STRINGLIST *m1, *m2;
-{
-  register int i, n, len1, len2;
-
-  if (m1 == 0)
-    {
-      m1 = copy_stringlist (m2);
-      return m1;
-    }
-
-  len1 = m1->list_len;
-  len2 = m2 ? m2->list_len : 0;
-
-  if (len2)
-    {
-      m1 = realloc_stringlist (m1, len1 + len2 + 1);
-      for (i = 0, n = len1; i < len2; i++, n++)
-	m1->list[n] = STRDUP (m2->list[i]);
-      m1->list[n] = (char *)NULL;
-      m1->list_len = n;
-    }
-
-  return m1;
-}
 
 static int
 shouldexp_filterpat (s)
@@ -391,7 +297,7 @@ filter_stringlist (sl, filterpat, text)
   ret = alloc_stringlist (sl->list_size);
   for (i = 0; i < sl->list_len; i++)
     {
-      m = fnmatch (t, sl->list[i], FNMATCH_EXTFLAG);
+      m = strmatch (t, sl->list[i], FNMATCH_EXTFLAG);
       if ((not && m == FNM_NOMATCH) || (not == 0 && m != FNM_NOMATCH))
 	free (sl->list[i]);
       else
@@ -405,55 +311,8 @@ filter_stringlist (sl, filterpat, text)
   return ret;
 }
 
-STRINGLIST *
-prefix_suffix_stringlist (sl, prefix, suffix)
-     STRINGLIST *sl;
-     char *prefix, *suffix;
-{
-  int plen, slen, tlen, llen, i;
-  char *t;
-
-  if (sl == 0 || sl->list == 0 || sl->list_len == 0)
-    return sl;
-
-  plen = STRLEN (prefix);
-  slen = STRLEN (suffix);
-
-  if (plen == 0 && slen == 0)
-    return (sl);
-
-  for (i = 0; i < sl->list_len; i++)
-    {
-      llen = STRLEN (sl->list[i]);
-      tlen = plen + llen + slen + 1;
-      t = xmalloc (tlen + 1);
-      if (plen)
-	strcpy (t, prefix);
-      strcpy (t + plen, sl->list[i]);
-      if (slen)
-	strcpy (t + plen + llen, suffix);
-      free (sl->list[i]);
-      sl->list[i] = t;
-    }
-
-  return (sl);	 
-}
-   
-void
-print_stringlist (sl, prefix)
-     STRINGLIST *sl;
-     char *prefix;
-{
-  register int i;
-
-  if (sl == 0)
-    return;
-  for (i = 0; i < sl->list_len; i++)
-    printf ("%s%s\n", prefix ? prefix : "", sl->list[i]);
-}
-
-/* Turn an array of strings returned by completion_matches into a STRINGLIST.
-   This understands how completion_matches sets matches[0] (the lcd of the
+/* Turn an array of strings returned by rl_completion_matches into a STRINGLIST.
+   This understands how rl_completion_matches sets matches[0] (the lcd of the
    strings in the list, unless it's the only match). */
 STRINGLIST *
 completions_to_stringlist (matches)
@@ -491,21 +350,21 @@ it_init_aliases (itp)
      ITEMLIST *itp;
 {
 #ifdef ALIAS
-  alias_t **aliases;
+  alias_t **alias_list;
   register int i, n;
   STRINGLIST *sl;
 
-  aliases = all_aliases ();
-  if (aliases == 0)
+  alias_list = all_aliases ();
+  if (alias_list == 0)
     {
       itp->slist = (STRINGLIST *)NULL;
       return 0;
     }
-  for (n = 0; aliases[n]; n++)
+  for (n = 0; alias_list[n]; n++)
     ;
   sl = alloc_stringlist (n+1);
   for (i = 0; i < n; i++)
-    sl->list[i] = STRDUP (aliases[i]->name);
+    sl->list[i] = STRDUP (alias_list[i]->name);
   sl->list[n] = (char *)NULL;
   sl->list_size = sl->list_len = n;
   itp->slist = sl;
@@ -554,7 +413,7 @@ it_init_bindings (itp)
   STRINGLIST *sl;
 
   /* rl_funmap_names allocates blist, but not its members */
-  blist = rl_funmap_names ();
+  blist = (char **)rl_funmap_names ();	/* XXX fix const later */
   sl = alloc_stringlist (0);
   sl->list = blist;
   sl->list_size = 0;
@@ -570,7 +429,6 @@ it_init_builtins (itp)
      ITEMLIST *itp;
 {
   STRINGLIST *sl;
-  char **list;
   register int i, n;
 
   sl = alloc_stringlist (num_shell_builtins);
@@ -588,7 +446,6 @@ it_init_enabled (itp)
      ITEMLIST *itp;
 {
   STRINGLIST *sl;
-  char **list;
   register int i, n;
 
   sl = alloc_stringlist (num_shell_builtins);
@@ -608,7 +465,6 @@ it_init_disabled (itp)
      ITEMLIST *itp;
 {
   STRINGLIST *sl;
-  char **list;
   register int i, n;
 
   sl = alloc_stringlist (num_shell_builtins);
@@ -661,7 +517,7 @@ it_init_joblist (itp, jstate)
 {
 #if defined (JOB_CONTROL)
   STRINGLIST *sl;
-  register int i, n;
+  register int i;
   register PROCESS *p;
   char *s, *t;
   JOB_STATE js;
@@ -683,7 +539,7 @@ it_init_joblist (itp, jstate)
 	  t = strpbrk (s, " \t\n");
 	  if (t)
 	    *t = '\0';
-	sl->list[sl->list_len++] = s;
+	  sl->list[sl->list_len++] = s;
 	}
     }
   itp->slist = sl;
@@ -789,7 +645,7 @@ it_init_shopts (itp)
 static STRINGLIST *
 gen_matches_from_itemlist (itp, text)
      ITEMLIST *itp;
-     char *text;
+     const char *text;
 {
   STRINGLIST *ret, *sl;
   int tlen, i, n;
@@ -802,6 +658,8 @@ gen_matches_from_itemlist (itp, text)
       if ((itp->flags & LIST_INITIALIZED) == 0)
 	initialize_itemlist (itp);
     }
+  if (itp->slist == 0)
+    return ((STRINGLIST *)NULL);
   ret = alloc_stringlist (itp->slist->list_len+1);
   sl = itp->slist;
   tlen = STRLEN (text);
@@ -814,11 +672,11 @@ gen_matches_from_itemlist (itp, text)
   return ret;
 }
 
-/* A wrapper for filename_completion_function that dequotes the filename
+/* A wrapper for rl_filename_completion_function that dequotes the filename
    before attempting completions. */
 static char *
 pcomp_filename_completion_function (text, state)
-     char *text;
+     const char *text;
      int state;
 {
   static char *dfn;	/* dequoted filename */
@@ -831,13 +689,13 @@ pcomp_filename_completion_function (text, state)
       if (rl_filename_dequoting_function)
 	{
 	  qc = (text[0] == '"' || text[0] == '\'') ? text[0] : 0;
-	  dfn = (*rl_filename_dequoting_function) (text, qc);
+	  dfn = (*rl_filename_dequoting_function) ((char *)text, qc);
 	}
       else
 	dfn = savestring (text);
     }
 
-  return (filename_completion_function (dfn, state));
+  return (rl_filename_completion_function (dfn, state));
 }
 
 #define GEN_COMPS(bmap, flag, it, text, glist, tlist) \
@@ -845,8 +703,11 @@ pcomp_filename_completion_function (text, state)
     if (bmap & flag) \
       { \
 	tlist = gen_matches_from_itemlist (it, text); \
-	glist = append_stringlist (glist, tlist); \
-	free_stringlist (tlist); \
+	if (tlist) \
+	  { \
+	    glist = append_stringlist (glist, tlist); \
+	    free_stringlist (tlist); \
+	  } \
       } \
   } while (0)
 
@@ -854,7 +715,7 @@ pcomp_filename_completion_function (text, state)
   do { \
     if (bmap & flag) \
       { \
-	cmatches = completion_matches (text, func); \
+	cmatches = rl_completion_matches (text, func); \
 	tlist = completions_to_stringlist (cmatches); \
 	glist = append_stringlist (glist, tlist); \
 	free_array (cmatches); \
@@ -867,10 +728,10 @@ pcomp_filename_completion_function (text, state)
 static STRINGLIST *
 gen_action_completions (cs, text)
      COMPSPEC *cs;
-     char *text;
+     const char *text;
 {
   STRINGLIST *ret, *tmatches;
-  char **cmatches;	/* from completion_matches ... */
+  char **cmatches;	/* from rl_completion_matches ... */
   unsigned long flags;
 
   ret = tmatches = (STRINGLIST *)NULL;
@@ -896,7 +757,8 @@ gen_action_completions (cs, text)
 
   GEN_XCOMPS(flags, CA_COMMAND, text, command_word_completion_function, cmatches, ret, tmatches);
   GEN_XCOMPS(flags, CA_FILE, text, pcomp_filename_completion_function, cmatches, ret, tmatches);
-  GEN_XCOMPS(flags, CA_USER, text, username_completion_function, cmatches, ret, tmatches);
+  GEN_XCOMPS(flags, CA_USER, text, rl_username_completion_function, cmatches, ret, tmatches);
+  GEN_XCOMPS(flags, CA_GROUP, text, bash_groupname_completion_function, cmatches, ret, tmatches);
 
   /* And lastly, the special case for directories */
   if (flags & CA_DIRECTORY)
@@ -919,10 +781,9 @@ gen_action_completions (cs, text)
 static STRINGLIST *
 gen_globpat_matches (cs, text)
       COMPSPEC *cs;
-      char *text;
+      const char *text;
 {
   STRINGLIST *sl;
-  char *t;
 
   sl = alloc_stringlist (0);
   sl->list = glob_filename (cs->globpat);
@@ -938,7 +799,7 @@ gen_globpat_matches (cs, text)
 static STRINGLIST *
 gen_wordlist_matches (cs, text)
      COMPSPEC *cs;
-     char *text;
+     const char *text;
 {
   WORD_LIST *l, *l2;
   STRINGLIST *sl;
@@ -965,7 +826,7 @@ gen_wordlist_matches (cs, text)
   for (nw = 0, l = l2; l; l = l->next)
     {
       if (tlen == 0 || STREQN (l->word->word, text, tlen))
-        sl->list[nw++] = STRDUP (l->word->word);
+	sl->list[nw++] = STRDUP (l->word->word);
     }
   sl->list[sl->list_len = nw] = (char *)NULL;
 
@@ -999,7 +860,7 @@ bind_compfunc_variables (line, ind, lwords, cw, exported)
      WORD_LIST *lwords;
      int cw, exported;
 {
-  char ibuf[32];
+  char ibuf[INT_STRLEN_BOUND(int) + 1];
   char *value;
   SHELL_VAR *v;
 
@@ -1009,7 +870,7 @@ bind_compfunc_variables (line, ind, lwords, cw, exported)
   if (v && exported)
     VSETATTR(v, att_exported);
 
-  value = inttostr (ind, ibuf, 32);
+  value = inttostr (ind, ibuf, sizeof(ibuf));
   v = bind_int_variable ("COMP_POINT", value);
   if (v && exported)
     VSETATTR(v, att_exported);
@@ -1020,7 +881,7 @@ bind_compfunc_variables (line, ind, lwords, cw, exported)
     {
 #ifdef ARRAY_VARS
       v = bind_comp_words (lwords);
-      value = inttostr (cw, ibuf, 32);
+      value = inttostr (cw, ibuf, sizeof(ibuf));
       bind_int_variable ("COMP_CWORD", value);
 #endif
     }
@@ -1056,7 +917,7 @@ unbind_compfunc_variables (exported)
 static WORD_LIST *
 build_arg_list (cmd, text, lwords, ind)
      char *cmd;
-     char *text;
+     const char *text;
      WORD_LIST *lwords;
      int ind;
 {
@@ -1099,7 +960,8 @@ build_arg_list (cmd, text, lwords, ind)
 static STRINGLIST *
 gen_shell_function_matches (cs, text, line, ind, lwords, nw, cw)
      COMPSPEC *cs;
-     char *text, *line;
+     const char *text;
+     char *line;
      int ind;
      WORD_LIST *lwords;
      int nw, cw;
@@ -1117,8 +979,8 @@ gen_shell_function_matches (cs, text, line, ind, lwords, nw, cw)
   f = find_function (funcname);
   if (f == 0)
     {
-      fprintf (stderr, "gen_shell_function_matches: function `%s' not found\n", funcname);
-      ding ();
+      internal_error ("completion: function `%s' not found", funcname);
+      rl_ding ();
       rl_on_new_line ();
       return ((STRINGLIST *)NULL);
     }
@@ -1177,7 +1039,8 @@ gen_shell_function_matches (cs, text, line, ind, lwords, nw, cw)
 static STRINGLIST *
 gen_command_matches (cs, text, line, ind, lwords, nw, cw)
      COMPSPEC *cs;
-     char *text, *line;
+     const char *text;
+     char *line;
      int ind;
      WORD_LIST *lwords;
      int nw, cw;
@@ -1198,14 +1061,14 @@ gen_command_matches (cs, text, line, ind, lwords, nw, cw)
   cmdsize += 2;
 
   /* allocate the string for the command and fill it in. */
-  cscmd = xmalloc (cmdsize + 1);
+  cscmd = (char *)xmalloc (cmdsize + 1);
 
   strcpy (cscmd, cs->command);			/* $0 */
   cmdlen = n;
   cscmd[cmdlen++] = ' ';
   for (cl = cmdlist->next; cl; cl = cl->next)   /* $1, $2, $3, ... */
     {
-      t = single_quote (cl->word->word ? cl->word->word : "");
+      t = sh_single_quote (cl->word->word ? cl->word->word : "");
       n = strlen (t);
       RESIZE_MALLOCED_BUFFER (cscmd, cmdlen, n + 2, cmdsize, 64);
       strcpy (cscmd + cmdlen, t);
@@ -1272,21 +1135,21 @@ command_line_to_word_list (line, llen, sentinel, nwp, cwp)
 STRINGLIST *
 gen_compspec_completions (cs, cmd, word, start, end)
      COMPSPEC *cs;
-     char *cmd;
-     char *word;
+     const char *cmd;
+     const char *word;
      int start, end;
 {
   STRINGLIST *ret, *tmatches;
-  char *line, *lword;
+  char *line;
   int llen, nw, cw;
   WORD_LIST *lwords;
 
   debug_printf ("programmable_completions (%s, %s, %d, %d)", cmd, word, start, end);
-  debug_printf ("programmable_completions: %s -> 0x%x", cmd, (int)cs);
+  debug_printf ("programmable_completions: %s -> %p", cmd, cs);
   ret = gen_action_completions (cs, word);
   if (ret && progcomp_debug)
     {
-      debug_printf ("gen_action_completions (0x%x, %s) -->", (int)cs, word);
+      debug_printf ("gen_action_completions (%p, %s) -->", cs, word);
       print_stringlist (ret, "\t");
       rl_on_new_line ();
     }
@@ -1299,7 +1162,7 @@ gen_compspec_completions (cs, cmd, word, start, end)
 	{
 	  if (progcomp_debug)
 	    {
-	      debug_printf ("gen_globpat_matches (0x%x, %s) -->", (int)cs, word);
+	      debug_printf ("gen_globpat_matches (%p, %s) -->", cs, word);
 	      print_stringlist (tmatches, "\t");
 	      rl_on_new_line ();
 	    }
@@ -1316,7 +1179,7 @@ gen_compspec_completions (cs, cmd, word, start, end)
 	{
 	  if (progcomp_debug)
 	    {
-	      debug_printf ("gen_wordlist_matches (0x%x, %s) -->", (int)cs, word);
+	      debug_printf ("gen_wordlist_matches (%p, %s) -->", cs, word);
 	      print_stringlist (tmatches, "\t");
 	      rl_on_new_line ();
 	    }
@@ -1335,7 +1198,7 @@ gen_compspec_completions (cs, cmd, word, start, end)
       line = substring (rl_line_buffer, start, end);
       llen = end - start;
 
-      debug_printf ("command_line_to_word_list (%s, %d, %d, 0x%x, 0x%x)",
+      debug_printf ("command_line_to_word_list (%s, %d, %d, %p, %p)",
 		line, llen, rl_point - start, &nw, &cw);
       lwords = command_line_to_word_list (line, llen, rl_point - start, &nw, &cw);
       if (lwords == 0 && llen > 0)
@@ -1358,7 +1221,7 @@ gen_compspec_completions (cs, cmd, word, start, end)
 	{
 	  if (progcomp_debug)
 	    {
-	      debug_printf ("gen_shell_function_matches (0x%x, %s, 0x%x, %d, %d) -->", (int)cs, word, lwords, nw, cw);
+	      debug_printf ("gen_shell_function_matches (%p, %s, %p, %d, %d) -->", cs, word, lwords, nw, cw);
 	      print_stringlist (tmatches, "\t");
 	      rl_on_new_line ();
 	    }
@@ -1374,7 +1237,7 @@ gen_compspec_completions (cs, cmd, word, start, end)
 	{
 	  if (progcomp_debug)
 	    {
-	      debug_printf ("gen_command_matches (0x%x, %s, 0x%x, %d, %d) -->", (int)cs, word, lwords, nw, cw);
+	      debug_printf ("gen_command_matches (%p, %s, %p, %d, %d) -->", cs, word, lwords, nw, cw);
 	      print_stringlist (tmatches, "\t");
 	      rl_on_new_line ();
 	    }
@@ -1395,7 +1258,7 @@ gen_compspec_completions (cs, cmd, word, start, end)
       tmatches = filter_stringlist (ret, cs->filterpat, word);
       if (progcomp_debug)
 	{
-	  debug_printf ("filter_stringlist (0x%x, %s, %s) -->", ret, cs->filterpat, word);
+	  debug_printf ("filter_stringlist (%p, %s, %s) -->", ret, cs->filterpat, word);
 	  print_stringlist (tmatches, "\t");
 	  rl_on_new_line ();
 	}
@@ -1410,6 +1273,20 @@ gen_compspec_completions (cs, cmd, word, start, end)
   if (cs->prefix || cs->suffix)
     ret = prefix_suffix_stringlist (ret, cs->prefix, cs->suffix);
 
+  /* If no matches have been generated and the user has specified that
+      directory completion should be done as a default, call
+      gen_action_completions again to generate a list of matching directory
+      names. */
+  if ((ret == 0 || ret->list_len == 0) && (cs->options & COPT_DIRNAMES))
+    {
+      COMPSPEC *dummy;
+
+      dummy = alloc_compspec ();
+      dummy->actions = CA_DIRECTORY;
+      ret = gen_action_completions (dummy, word);
+      free_compspec (dummy);
+    }
+
   return (ret);
 }
 
@@ -1418,7 +1295,8 @@ gen_compspec_completions (cs, cmd, word, start, end)
    bound the command currently being completed in rl_line_buffer. */
 char **
 programmable_completions (cmd, word, start, end, foundp)
-     char *cmd, *word;
+     const char *cmd;
+     const char *word;
      int start, end, *foundp;
 {
   COMPSPEC *cs;
@@ -1441,9 +1319,10 @@ programmable_completions (cmd, word, start, end, foundp)
       return ((char **)NULL);
     }
 
-  /* Signal the caller that we found a COMPSPEC for this command. */
+  /* Signal the caller that we found a COMPSPEC for this command, and pass
+     back any meta-options associated with the compspec. */
   if (foundp)
-    *foundp = 1;
+    *foundp = 1|cs->options;
 
   ret = gen_compspec_completions (cs, cmd, word, start, end);
 
